/* Copyright (c) 2012-2014, The Ohio State University. All rights
 * reserved.
 *
 * This file is part of the MVAPICH2 software package developed by the
 * team members of The Ohio State University's Network-Based Computing
 * Laboratory (NBCL), headed by Professor Dhabaleswar K. (DK) Panda.
 *
 * For detailed copyright and licensing information, please refer to the
 * copyright file COPYRIGHT in the top level MVAPICH2 directory.
 *
 */
#ifndef _UCR_H_
#define _UCR_H_

#define UCR_MSG_REMOTE_PIN_REQ      65
#define UCR_MSG_REMOTE_PIN_RSP      66

#define UCR_MSG_REMOTE_UNPIN_REQ    67
#define UCR_MSG_REMOTE_UNPIN_RSP    68

typedef void ucr_mkey_t;
typedef int fp_ucr_am_short_handler_t(int src, int msg_id, void *msg, size_t msg_len);
typedef int fp_ucr_am_medium_handler_t(int src, int msg_id, void *hdr, size_t hdr_len, 
        void *msg, size_t msg_len);

/* init */
int ucr_init(int *argc, char ***argv,
        fp_ucr_am_short_handler_t fp_ucr_am_short_handler,
        fp_ucr_am_medium_handler_t fp_ucr_am_medium_handler);

/* get rank */
int ucr_get_pg_rank();

/* get size */
int ucr_get_pg_size();

/* check if a process in on the same node
 * returns -1 if not local */
int ucr_get_local_rank(int rank);

/*get number of proceses on the local node*/
int ucr_get_local_pg_size();

/* Establish connection with rank if unconnected */
int ucr_establish_connection(int rank);

/* send short active msg */
int ucr_send_short_am(int rank, int msg_id, 
        void *msg, size_t msg_len);

/* send medium active msg */
int ucr_send_medium_am(int rank, int msg_id,
        void *hdr, size_t hdr_len,
        void *msg, size_t msg_len);

/* register memory */
ucr_mkey_t *ucr_register_memory(void *addr, size_t len);

/* register memory (remote) */
ucr_mkey_t *ucr_register_memory_remote(int rank, void *addr, size_t len, uint32_t *p_rkey);

/* get mkey size */
int ucr_mkey_get_size();

/* unregister memory */
void ucr_unregister_memory(ucr_mkey_t *mkey);

/* serialize mkey */
void *ucr_serialize_mkey(ucr_mkey_t *mkey, int *num_bytes);

/* deserialize mkey */
ucr_mkey_t *ucr_deserialize_mkey(void *data, int num_bytes);

/* free deserialized key */
void ucr_free_deserialized_mkey(ucr_mkey_t *mkey);

/* select rail */
int ucr_select_rail(int rank);

/* get key */
uint32_t ucr_get_mr_key(int rank, int rail, ucr_mkey_t *mkey);

/* put */
int ucr_put(int rank, void *dst, void *src, size_t len,
        uint32_t rkey, uint32_t lkey, int rail, int *counter);

/* put */
int ucr_put_local_cmp(int rank, void *dst, void *src, size_t len,
        uint32_t rkey, uint32_t lkey, int rail, int *counter);

/* get */
int ucr_get(int rank, void *src, void *dst, size_t len,
        uint32_t rkey, uint32_t lkey, int rail, int *counter);

/* fetch-add */
int ucr_fetch_add(int rank,
        void *local_addr, uint32_t mr_lkey, 
        void *remote_addr, uint32_t mr_rkey, 
        uint64_t value, size_t len, 
        int rail, int *counter);

/* compare-swap */
int ucr_compare_swap(int rank, 
        void *local_addr, uint32_t mr_lkey, 
        void *remote_addr, uint32_t mr_rkey,
        uint64_t compare, uint64_t swap,
        size_t len, 
        int rail, int *counter);

/* barrier */
int ucr_barrier();

/* all-to-all */
int ucr_alltoall(void *src, size_t len, void *dest);

/* exchange */
int ucr_exchange(void *src, size_t len, void *dest);

/* probe */
int ucr_probe();

/* GASNet Bindings */

/* handlers */
typedef int fp_ucr_gasnet_am_short_handler_t(uint16_t token, uint8_t handler,
        int32_t *args, uint8_t numargs);

typedef int fp_ucr_gasnet_am_medium_handler_t(uint16_t token, uint8_t handler,
        int32_t *args, uint8_t numargs, void *buf, uint32_t buflen);

typedef int fp_ucr_gasnet_am_long_handler_t(uint16_t token, uint8_t handler,
        int32_t *args, uint8_t numargs, void *buf, uint32_t buflen);

typedef int fp_ucr_gasnet_am_system_handler_t(uint16_t token, uint8_t handler,
        int32_t *args, uint8_t numargs);

/* init */
int ucr_gasnet_init(int *argc, char ***argv,
        fp_ucr_gasnet_am_short_handler_t fp_ucr_gasnet_am_short_handler,
        fp_ucr_gasnet_am_medium_handler_t fp_ucr_gasnet_am_medium_handler,
        fp_ucr_gasnet_am_long_handler_t fp_ucr_gasnet_am_long_handler,
        fp_ucr_gasnet_am_system_handler_t fp_ucr_gasnet_am_system_handler);

/* short am */
int ucr_gasnet_short_am(uint16_t rank, uint16_t token, uint8_t handler, 
        va_list argptr, uint8_t numargs);

/* medium am */
int ucr_gasnet_medium_am(uint16_t rank, uint16_t token, uint8_t handler,
        void *source_addr, uint32_t nbytes, va_list argptr, uint8_t numargs);

/* long am */
int ucr_gasnet_long_am(uint16_t rank, uint16_t token, uint8_t handler,
        void *source_addr, uint32_t nbytes, uint32_t lkey, va_list argptr,
        uint8_t numargs, void *dest_addr, uint32_t rkey, 
        int *p_counter);

/* system am */
int ucr_gasnet_system_am(uint16_t rank, uint16_t token, uint8_t handler, 
        va_list argptr, uint8_t numargs);

/* get max_msg_sz */
uint32_t ucr_get_max_msg_sz();

/* UCR Collecives */
int ucr_coll_barrier(int PE_start, int logPE_stride, int PE_size);

int ucr_coll_broadcast(void *target, const void *source, size_t nelems,
		       int PE_root, int PE_start,
		       int logPE_stride, int PE_size);

int ucr_coll_collect(void *target, const void *source, size_t nelems,
		     int PE_start, int logPE_stride, int PE_size);

int ucr_coll_collectv(void *target, const void *source, size_t nelems,                                   
                      int *recvcounts, int *displs,                                                          
                      int PE_start, int logPE_stride, int PE_size);

int ucr_coll_reduce_all(void *target, void *source, int nreduce,
			MPI_Datatype datatype, MPI_Op op,
			int PE_start, int logPE_stride, int PE_size);

int ucr_coll_reduce(void *target, void *source, int nreduce, 
                    MPI_Datatype datatype, MPI_Op op, int root,
                    int PE_start, int logPE_stride, int PE_size);

int ucr_coll_scatter(void *target, void *source, int len, int root,
                     int PE_start, int logPE_stride, int PE_size);

int ucr_coll_alltoall(void *source, void *target, int len,
                      int PE_start, int logPE_stride, int PE_size);

int ucr_coll_gather(void *target, const void *source, size_t nelems,
                    int root, int PE_start, int logPE_stride, int PE_size);


/* finalize */
void ucr_finalize();

#endif
